(define (bst/insert! val tree)
  ; Space: Theta(log n) if balanced, up to Theta(n) otherwise
  ; Time : Theta(log n) if balanced, up to Theta(n) otherwise
  ;
  ; Recursive version (deferred set-left-subtree!)
  (check-binary-tree tree)
  (cond ((empty-binary-tree? tree)
         (make-node val *empty-binary-tree* *empty-binary-tree*))
        ((< val (node-value tree))
         (set-left-subtree! tree 
                            (bst/insert! val (left-subtree tree))))
        (else
         (set-right-subtree! tree 
                             (bst/insert! val (right-subtree tree))))))


; below are a set of increasingly "better" solutions that try to 
; use iteration and do so efficiently.


(define (bst/insert! val tree)
  ; Space: Theta(log n) if balanced, up to Theta(n) otherwise
  ; Time : Theta(log n) if balanced, up to Theta(n) otherwise
  ;
  ; looks iterative, but isn't (the tree return values count as deferred operations)
  (cond ((not (binary-tree? tree)) 
         (error "not a tree:" tree))
        
        ((empty-binary-tree? tree)
         (make-node val *empty-binary-tree* *empty-binary-tree*))
        
        ((< val (node-value tree))
         (if (empty-binary-tree? (left-subtree tree))
             (set-left-subtree! tree
                                (bst/insert! val (left-subtree tree)))
             (begin 
               (bst/insert! val (left-subtree tree))
               tree)))
        
        (else 
         (if (empty-binary-tree? (right-subtree tree))
             (set-right-subtree! tree
                                 (bst/insert! val (right-subtree tree)))
             (begin 
               (bst/insert! val (right-subtree tree))
               tree)))))


(define (bst/insert! val tree)
  ; Space: Theta(1) always
  ; Time : Theta(log n) if balanced, up to Theta(n) otherwise
  ;
  ; Okay, this one actually is iterative.  We create a handle: a cons cell that
  ; has a dummy car part and whose cdr part points to the subtree we're handling.
  ; We use the handle to simulate pass-by-reference (note: this is actually quite
  ; similar to the way C simulates pass-by-reference by using pointers).
  (define (iter handle)
    (cond 
      ; Here it's easiest to handle the error condition first
      ((not (binary-tree? (cdr handle))) 
       (error "not a tree:" (cdr handle)))
      
      ; special base case: create a new node if the original tree was empty.  Note:
      ; the handle will never be null (it's an extra cons cell that was tacked onto
      ; our tree), so it's always safe to set-cdr! to it.
      ((empty-binary-tree? (cdr handle))
       (set-cdr! handle (make-node val *empty-binary-tree* *empty-binary-tree*)))
      
      ; Need to go down the left subtree?
      ((< val (node-value (cdr handle)))
       (if (empty-binary-tree? (left-subtree (cdr handle)))
           ; we're about to walk off the tree to the left, so stop early and
           ; attach a new node
           (set-left-subtree! (cdr handle)
                              (make-node val *empty-binary-tree* *empty-binary-tree*))
           ; create a new handle and step down the left subtree
           (iter (cons '() (left-subtree (cdr handle))))))
      
      ; right subtree case looks similar to the left subtree...
      (else 
       (if (empty-binary-tree? (right-subtree (cdr handle)))
           (set-right-subtree! (cdr handle)
                               (make-node val *empty-binary-tree* *empty-binary-tree*))
           (iter (cons '() (right-subtree (cdr handle))))))))
  
  ; First, create and keep a handle to the root of our tree.  This allows us to
  ; handle null trees
  (let ((handle (cons '() tree)))
    ; Do the mutating insertion.  Its return value is undefined.
    (iter handle)
    (cdr handle)))


(define (bst/insert! val tree)
  ; Space: Theta(1) always
  ; Time : Theta(log n) if balanced, up to Theta(n) otherwise
  ;
  ; Here's another iterative version.  In this case, we avoid using the handle
  ; by handling the special empty tree case outside of the iterator.
  (define (iter tree)
    (cond 
      ((not (binary-tree? tree)) 
       (error "not a tree:" tree))
            
      ((< val (node-value tree))
       (if (empty-binary-tree? (left-subtree tree))
           (set-left-subtree! tree
                              (make-node val *empty-binary-tree* *empty-binary-tree*))
           (iter (left-subtree tree))))
      
      (else 
       (if (empty-binary-tree? (right-subtree tree))
           (set-right-subtree! tree
                               (make-node val *empty-binary-tree* *empty-binary-tree*))
           (iter (right-subtree tree))))))

  (if (empty-binary-tree? tree)
      ; keep iter simple by handling the difficult case of any empty tree out here
      (make-node val *empty-binary-tree* *empty-binary-tree*)
      (begin
        ; otherwise, do the simplified iteration
        (iter tree)
        ; and don't forget to return the final result since iter returns junk (not
        ; worrying about the return value is what made changing to an iterative
        ; procedure easier than it would have been)
        tree)))